#include "hourglass.h"

double MHZ = 0;
int num_threads = 0;
unsigned int duration;

struct thrd_data_t *thrd_data = NULL;
struct trace_rec_t *trace_recs = NULL;
long current_record;
int tracing;
cycle_time start, end;
int need_timers = 0;
double gap_us;
cycle_time GAP;

#ifdef GAP_HIST
int *gap_hist;
#endif

static void wait_for_all_threads (enum thrd_status_t new_status)
{
  int cnt = 0;
  int ready;
  int i;

  do {
    if (cnt++ > MAX_SPIN) {
      printf ("Hrm: waited too long for threads to reach state %d\n",
	      new_status);
      exit (-1);
    }
    usleep (SMALL_SLEEP);
    ready = 1;
    for (i=0; i<num_threads; i++) {
      if (thrd_data[i].status != new_status) {
	ready = 0;
      }
    }
  } while (!ready);
}

static void set_all_threads (enum thrd_status_t new_status)
{
  int i;

  for (i=0; i<num_threads; i++) {
    thrd_data[i].status = new_status;
  }
}

static void flush_records (void)
{
  int i, rec;

  for (i=0; i<num_threads; i++) {
    rec = thrd_data[i].reserved_record;
    if (rec < 0) {
      continue;
    }
    trace_recs[rec].id = i;
    trace_recs[rec].start = thrd_data[i].exec_start;
    trace_recs[rec].end = thrd_data[i].last_exec;
  }
}
  
int main (int argc, char *argv[])
{
  int i, retval;

  // time conversion functions won't work until this runs
  MHZ = find_mhz ();

  parse_args (argc, argv);

  // fixme
  gap_us = 2.0;
  duration = 10000000;
  
  if (tracing) {
    GAP = us_to_cycle (gap_us);
    printf ("max gap is %lld cycles\n", GAP);
  }

  if (num_threads == 0) {
    printf ("Hrm: internal error\n");
    exit (-1);
  }

  fprintf (stderr, "Hourglass %s running %d thread%s on a %f MHz processor\n",
	   VERSION, 
	   num_threads, 
	   (num_threads == 1) ? "" : "s",
	   MHZ);

  lock_pages ();
  set_priority (MAIN_PRI);
  if (need_timers) {
    init_timers ();
  }

  printf ("this test will last for %f seconds\n", 
	  us_to_sec ((double)duration));

  set_all_threads (INIT);

  for (i=0; i<num_threads; i++) {
    pthread_attr_t attr;
    
    init_attr (&attr, thrd_data[i].pri);
    retval = pthread_create (&thrd_data[i].thrd,
			     &attr,
			     thrd_func,
			     (void *)i);
    if (retval != 0) {
      printf ("Hrm: error creating thread %d\n", i);
      exit (-1);
    }
  }

  // wait_for_all_threads (READY);
  set_all_threads (GO);
  start = rdtsc();

  usleep (duration);

  set_all_threads (STOP);
  wait_for_all_threads (DONE);
  end = rdtsc();
  if (tracing) {
    flush_records ();
  }
  set_priority (NORMAL);
  if (need_timers) {
    deinit_timers ();
  }
  print_all_results ();

  return 0;
} 
